/* tib2xxu.c - TI-Base to .??u OS file converter
 * Copyright (C) 2005 Olivier Armand - <ola dot e-ml at wanadoo dot fr>
 * 
 * 2005-03-20:
 *  * Created
 */

#define WINDOWS /* define to allow using Win32 API */

/* *NIX compatibility for case-unsensitive comparision */
#ifdef WINDOWS
#define STRCMPI strcmpi
#else
#define STRCMPI strcasecmp
#endif

#ifdef WINDOWS
# include <windows.h>
#endif

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#ifdef WINDOWS
#define TIB_PROMPT  "TIB file (*.tib)\0*.tib\0"
#define XXU_PROMPT  ".??u file (*.89u;*.9xu;*.v2u)\0*.89u;*.9xu;*.v2u\0"
static char strInFile[MAX_PATH] = "";
static char strOutFile[MAX_PATH] = "";
static OPENFILENAME OpenFileName = { sizeof (OPENFILENAME), 0, 0,
  TIB_PROMPT,
  NULL, 0, 0, strInFile, MAX_PATH, NULL, 0, NULL,
  "Select the TI-Base file...", OFN_FILEMUSTEXIST, 0, 0, NULL, 0, NULL, NULL
};
static OPENFILENAME SaveFileName = { sizeof (OPENFILENAME), 0, 0,
  XXU_PROMPT, NULL, 0,
  0, strOutFile, MAX_PATH, NULL, 0, NULL, "Save the TI-Update file as... (choose a file extension!)",
  OFN_OVERWRITEPROMPT, 0, 0, NULL, 0, NULL, NULL
};
#endif /* WINDOWS */

static const unsigned char XxuHeaderTemplate[] = {
  '*', '*', 'T', 'I', 'F', 'L', '*', '*',
  0x03, 0x00, 0x00, 0x00, /* ??? AMS major version number? */
  0x01, 0x01, 0x20, 0x05, /* BCD coded date (happy new year) */
  0x08, /* Name length */
  'b', 'a', 's', 'e', 'c', 'o', 'd', 'e', /* Name */
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, /* Nothing */
  0x00, /* Calculator type (to be filled): 0x98 for TI-89/Titanium, 0x88 for TI-92+/V200 */
  0x23, /* vartype = AMS */
  0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* Nothing */
  0x00, 0x00, 0x00, 0x00, /* OS size (to be filled) */
};


static unsigned char *indata = NULL;
static FILE *infile, *outfile;

static const char *wrong_outext_err = "Wrong output file extension.";


#ifdef WINDOWS
void
FatalWindows (LPCTSTR msg)
{
  MessageBox (NULL, msg, "Error...", MB_OK | MB_ICONERROR);
}
#endif /* WINDOWS */

void
ErrorMsg (const char *msg)
{
#ifdef WINDOWS
  FatalWindows (msg);
#else
  puts (msg);
#endif
}

void
Fatal (const char *msg)
{
  ErrorMsg (msg);
  exit(1);
}

void
FatalMsg (const char *msg)
{
#ifdef WINDOWS
  FatalWindows ((LPCTSTR) _strerror (msg));
#else
  perror (msg);
#endif
  exit (0);
}

void
xfwrite (const void *src, size_t n)
{
  if (fwrite (src, sizeof (char), n, outfile) != n)
    Fatal ("Error while writing to the output file.");
}

/* Return the calculator type identifier, or 0 if the file extension
 * is invalid. */
unsigned char
GetCalctype(const char *filename)
{
      char *inext = strrchr(filename, '.');
      if (inext == NULL)
    	return 0;
      else if (!strcmp (inext, ".89u"))
        return 0x98;
      else if (!strcmp (inext, ".v2u") || !strcmp(inext, ".9xu"))
        return 0x88;
      else
        return 0;
}


#ifndef WINDOWS
int
main (int argc, char *argv[])
#else /* WINDOWS */
int WINAPI
WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine,
	 int nCmdShow)
#endif /* WINDOWS */
{
  char *inname;
  char *outname;
  char *inext;
  long insize;
  unsigned char calctype;

  infile = outfile = NULL;

  /* Get the infile name */
#ifndef WINDOWS
  if (argc != 3) {
    printf ("Usage: %s infile.tib outfile.??u\n", argv[0]);
    exit (0);
  }
  inname = argv[1];
  outname = argv[2];
  inext = strrchr(inname, '.');
  if (inext == NULL || strcmp (".tib", inext))
   Fatal ("Input file is not a .tib file.");
  calctype = GetCalctype(outname);
  if (calctype == 0)
    Fatal (wrong_outext_err);
  
#else /* WINDOWS */
  while (1) {
  	  strInFile[0] = '\0';
	  if (!GetOpenFileName (&OpenFileName))
	    exit (0);
	  inname = strInFile;
      strcpy (strOutFile, strInFile);
      /* Is the extension ok? */
      outname = strOutFile;
      inext = strrchr(outname, '.');
      if (inext == NULL || strcmp (".tib", inext)) {
        FatalWindows ("Please select a .tib file.");
        continue;
      }
      strcpy(inext, ".??u");
      break;
  } /* while (1) */
  while (1) {

    if (!GetSaveFileName (&SaveFileName))
      exit (0);
    /* Find the calculator type for the xxu header */
    calctype = GetCalctype(outname);
    if (calctype == 0)
      FatalWindows (wrong_outext_err);
    else
   	  break;
  } /* while (1) */
#endif /* WINDOWS */

  /* Read the infile */
  if (!(infile = fopen (inname, "rb")))
    FatalMsg (inname);
  fseek (infile, 0, SEEK_END);
  insize = ftell (infile);
  rewind (infile);
  indata = malloc (sizeof(XxuHeaderTemplate) + insize);
  if (!indata || fread (indata + sizeof(XxuHeaderTemplate), sizeof (char), insize, infile) != (size_t)insize)
    FatalMsg ("Error while reading the input file.");
  fclose (infile);
  memcpy(indata, XxuHeaderTemplate, sizeof(XxuHeaderTemplate)); /* Copy the template header */

  /* Correct the calculator type in the header */
  indata[0x30] = calctype;
  /* Write the OS size in big endian in the header */
  indata[0x4A] = (unsigned char)insize;
  indata[0x4B] = (unsigned char)(insize >> 8);
  indata[0x4C] = (unsigned char)(insize >> 16);
  indata[0x4D] = (unsigned char)(insize >> 24);

  /* Prepare the output file */
  if (!(outfile = fopen (outname, "wb")))
    FatalMsg (outname);
  xfwrite (indata, sizeof(XxuHeaderTemplate) + insize); /* Copy the OS */
  free (indata);
#ifdef WINDOWS
  MessageBox (NULL, "TIB file converted successfully.", "Tib2xxu: Success!",
	      MB_OK | MB_ICONINFORMATION);
#else
  puts ("TIB file converted successfully.");
#endif
  exit (0);
}

